import functools

import pandas as pd
import numpy as np

from .. import accelerometer as accel
from .. import extensions as ext
from .. import mhealth_format as mh

INERTIAL_FEATURE_NAME_PREFIX = accel.STAT_FEATURE_NAME_PREFIX + \
    accel.SPECTRUM_FEATURE_NAME_PREFIX + \
    accel.ACTIVATION_FEATURE_NAME_PREFIX + \
    accel.ORIENTATION_FEATURE_NAME_PREFIX


def assemble_fv_names(selected=INERTIAL_FEATURE_NAME_PREFIX, use_vm=True, num_of_axes=3):
    fv_names = []
    for prefix in selected:
        if prefix in accel.ORIENTATION_FEATURE_NAME_PREFIX:
            fv_names += [f"{prefix}_{i}" for i in [0, 1, 2]]
        else:
            if use_vm:
                fv_names += [f"{prefix}_0"]
            else:
                fv_names += [f"{prefix}_{i}" for i in range(num_of_axes)]
    return fv_names


def single_triaxial(raw_df, sr, st, et, subwin_secs=2, ori_unit='rad', activation_threshold=0.2, use_vm=True, selected=INERTIAL_FEATURE_NAME_PREFIX):

    result = {
        mh.TIMESTAMP_COL: [st],
        mh.START_TIME_COL: [st],
        mh.STOP_TIME_COL: [et]
    }

    num_samples = round((et - st).total_seconds() * sr)

    if raw_df.shape[0] == 0 or raw_df.dropna().shape[0] < num_samples * 0.8:
        # TO BE IMPROVED
        # Now input raw data should include no less than 80% of the whole window
        fv_names = assemble_fv_names(selected=selected, use_vm=use_vm)
        for name in fv_names:
            result[name] = [np.nan]
        feature_names = fv_names
    else:
        # Prepare parameters
        subwin_samples = int(subwin_secs * sr)

        # Unify input matrix
        X = ext.numpy.atleast_float_2d(raw_df.values[:, 1:4])
        # fill nan at first, nan will be filled by spline interpolation
        X = ext.numpy.mutate_nan(X)

        # Smoothing input raw data
        if use_vm:
            X_vm = ext.numpy.vector_magnitude(X)
            X_vm_filtered = ext.numpy.butterworth(
                X_vm, sr=sr, cut_offs=20, order=4, filter_type='low')

        X_filtered = ext.numpy.butterworth(X, sr=sr, cut_offs=20,
                                           order=4, filter_type='low')

        ts_feature_funcs = [
            functools.partial(accel.stat_features, selected=selected),
            functools.partial(accel.spectrum_features,
                              sr=sr, n=1, selected=selected),
            functools.partial(accel.activation_features,
                              threshold=activation_threshold, selected=selected)
        ]
        orient_feature_funcs = [
            functools.partial(accel.orientation_features,
                              subwin_samples=subwin_samples, unit=ori_unit, selected=selected)
        ]
        if use_vm:
            X_for_ts = X_vm_filtered
        else:
            X_for_ts = X_filtered

        feature_names = []
        for func in ts_feature_funcs:
            values, names = func(X_for_ts)
            if values is None:
                continue
            for value, name in zip(values.transpose(), names):
                result[name] = value.tolist()
                feature_names.append(name)

        for func in orient_feature_funcs:
            values, names = func(X_filtered)
            if values is None:
                continue
            for value, name in zip(values.transpose(), names):
                result[name] = value.tolist()
                feature_names.append(name)
    result = pd.DataFrame.from_dict(result)
    return result, feature_names



from tqdm import tqdm
import joblib


class ProgressParallel(joblib.Parallel):
    def __init__(self, use_tqdm=True, total=None, *args, **kwargs):
        self._use_tqdm = use_tqdm
        self._total = total
        super().__init__(*args, **kwargs)

    def __call__(self, *args, **kwargs):
        with tqdm(disable=not self._use_tqdm, total=self._total) as self._pbar:
            return joblib.Parallel.__call__(self, *args, **kwargs)

    def print_progress(self):
        if self._total is None:
            self._pbar.total = self.n_dispatched_tasks
        self._pbar.n = self.n_completed_tasks
        self._pbar.refresh()



"""
Demonstration of the usage of arus.core.pipeline
=====================================================================

This example shows how to start, stop, pause a pipeline.
"""

from arus.core.pipeline import Pipeline
import arus
import datetime as dt
import pandas as pd
import time
import logging


def _pipeline_test_processor(chunk_list, **kwargs):
    import pandas as pd
    result = {'NAME': [],
              'START_TIME': [], 'STOP_TIME': []}
    for data, st, et, prev_st, prev_et, name in chunk_list:
        result['NAME'].append(name)
        result['START_TIME'].append(data.iloc[0, 0])
        result['STOP_TIME'].append(data.iloc[-1, 0])
    result = pd.DataFrame.from_dict(result)
    return result


if __name__ == "__main__":
    window_size = 2
    start_time = dt.datetime.now()

    gr = arus.generator.RandomAccelDataGenerator(
        sr=80, grange=8, sigma=1, buffer_size=100)
    seg = arus.segmentor.SlidingWindowSegmentor(window_size)
    stream = arus.Stream(gr, seg, name='stream-1')

    pipeline = Pipeline(max_processes=2, scheduler='processes')
    pipeline.add_stream(stream)
    pipeline.set_processor(_pipeline_test_processor)

    # connect, there will be no incoming data, get_iteratnor will be blocking
    pipeline.connect()
    results = []
    count_none = 0
    for result, st, et, prev_st, prev_et, name in pipeline.get_iterator(timeout=0.1):
        if result is not None:
            print('Connect is not working')
            assert False
        else:
            count_none = count_none + 1
        if count_none == 50:
            break
    print('Connected yet no data is coming')
    pipeline.stop()

    # connect, wait for 3 seconds and then start processing, get_iterator will be blocking for 3 + 4 seconds until receiving the first window
    pipeline.connect()
    connect_time = pd.Timestamp(datetime.now())
    print('Connect time: ' + str(connect_time))
    count_none = 0
    for result, st, et, prev_st, prev_et, name in pipeline.get_iterator(timeout=0.1):
        if result is not None:
            first_block_time = pd.Timestamp(datetime.now())
            print('First block time: ' + str(first_block_time) + ', ' +
                  str(first_block_time.timestamp() - connect_time.timestamp()) + ' seconds since connection.')
            print('First block st: ' + str(st))
            break
        else:
            count_none = count_none + 1
        if count_none == 30:
            process_time = pd.Timestamp(datetime.now())
            pipeline.process(start_time=process_time)
            print('Process time: ' + str(process_time) + ', ' +
                  str(process_time.timestamp() - connect_time.timestamp()) + ' seconds since connection.')
    print('Stop')
    pipeline.stop()

    # connect and process for 4 seconds and then pause for 2 seconds and then process again for 4 seconds
    st = datetime.now()
    pipeline.start(process_start_time=st)
    print('Start at: ' + str(st))
    count = 0
    count_none = 0
    restarted = False
    stt = time.time() + 100000
    for result, st, et, prev_st, prev_et, name in pipeline.get_iterator(timeout=0.1):
        if result is not None:
            count = count + 1
            print('Blocks at: ' + str(st))
            if count == 2:
                print('Pause')
                pipeline.pause()
                stt = time.time()
            if count == 4:
                break
        if time.time() - stt >= 2 and not restarted:
            restarted = True
            ts = dt.datetime.now()
            process_time = pd.Timestamp(ts)
            pipeline.process(start_time=process_time)
            print('Start again at:' + str(ts))

    print('Stop')
    pipeline.stop()



"""
Demonstration of the usage of arus.models.muss.get_inference_pipeline
=====================================================================

The pipeline uses multiple sensor generator streams.
"""

from arus.models.muss import MUSSModel
from arus.testing import load_test_data
import pandas as pd
import arus
import datetime as dt


def train_test_classifier(muss):
    feature_filepath, _ = load_test_data(file_type='mhealth', sensor_type='feature',
                                         file_num='single', exception_type='multi_placements')
    class_filepath, _ = load_test_data(file_type='mhealth', sensor_type='class_labels',
                                       file_num='single', exception_type='multi_tasks')
    feature_set = pd.read_csv(
        feature_filepath, infer_datetime_format=True, parse_dates=[0, 1, 2])
    dw_feature_set = feature_set.loc[feature_set['SENSOR_PLACEMENT'] == 'DW', [
        'HEADER_TIME_STAMP', 'START_TIME', 'STOP_TIME'] + muss.get_feature_names()]
    da_feature_set = feature_set.loc[feature_set['SENSOR_PLACEMENT'] == 'DA', [
        'HEADER_TIME_STAMP', 'START_TIME', 'STOP_TIME'] + muss.get_feature_names()]
    combined_feature_set, feature_names = muss.combine_features(
        dw_feature_set, da_feature_set, placement_names=['DW', 'DA'], group_col=None)
    class_set = pd.read_csv(
        class_filepath, infer_datetime_format=True, parse_dates=[0, 1, 2])
    class_set = class_set[['HEADER_TIME_STAMP',
                           'START_TIME', 'STOP_TIME', 'MUSS_3_POSTURES']]
    input_feature, input_class = muss.sync_feature_and_class(
        combined_feature_set, class_set)
    filtered_feature, filtered_class = muss.remove_classes(
        input_feature, input_class, class_col='MUSS_3_POSTURES', classes_to_remove=['Unknown', 'Transition'])
    model = muss.train_classifier(
        filtered_feature, filtered_class, class_col='MUSS_3_POSTURES', feature_names=feature_names, placement_names=['DW', 'DA'])
    return model


def prepare_streams():
    window_size = 12.8
    gr1 = arus.generator.RandomAccelDataGenerator(
        sr=50, grange=8, sigma=1, buffer_size=100)
    seg1 = arus.segmentor.SlidingWindowSegmentor(window_size)
    stream1 = arus.Stream(gr1, seg1, name='DW')
    gr2 = arus.generator.RandomAccelDataGenerator(
        sr=50, grange=4, sigma=2, buffer_size=100)
    seg2 = arus.segmentor.SlidingWindowSegmentor(window_size)
    stream2 = arus.Stream(gr2, seg2, name='DA')
    return stream1, stream2


if __name__ == "__main__":
    start_time = dt.datetime.now()
    muss = MUSSModel()
    model = train_test_classifier(muss)
    stream1, stream2 = prepare_streams()
    muss_pipeline = muss.get_inference_pipeline(
        stream1, stream2, model=model, scheduler='processes', max_processes=2, DW={'sr': 50}, DA={'sr': 50})
    muss_pipeline.start(start_time=start_time)
    i = 0
    for data, _, _, _, _, name in muss_pipeline.get_iterator():
        i = i + 1
        print(data)
        if i == 5:
            break
    muss_pipeline.stop()



"""
Module of extension functions to be applied to numpy objects (e.g., Arrays)

Author: Qu Tang

Date: 2020-02-03

License: see LICENSE file
"""

import numpy as np
import numpy.linalg as la
from scipy import interpolate
from scipy import signal as sp_signal


def atleast_scalar(arg):
    if np.isscalar(arg):
        result = np.float64(arg)
    else:
        result = np.array(arg, dtype='float64')
    return result


def atleast_float_2d(arr):
    arr = np.float64(arr)
    arr = np.atleast_2d(arr)
    return arr


def mutate_nan(X):
    X = atleast_float_2d(X)
    X_new = np.apply_along_axis(_fill_nan_1d, axis=0, arr=X)
    return X_new


def _fill_nan_1d(y):
    xnew = range(len(y))
    x, y = _remove_nan_1d(y)
    if len(x) < 3:
        ynew = np.interp(xnew, x, y)
    else:
        s = interpolate.InterpolatedUnivariateSpline(x, y)
        ynew = s(xnew)
    return ynew


def regularize_sr(t, X, sr, st=None, et=None):
    X = atleast_float_2d(X)
    st = st or t[0]
    et = et or t[-1]
    total_seconds = et - st
    new_t = np.linspace(st, et, num=int(
        np.floor(total_seconds * sr)), endpoint=False)
    new_X = np.apply_along_axis(
        _regularize_sr, axis=0, arr=X, t=t, new_t=new_t)
    return new_t, new_X


def _regularize_sr(y, t, new_t):
    f = interpolate.InterpolatedUnivariateSpline(t, y, k=3)
    new_y = f(new_t)
    return new_y


def _remove_nan_1d(y):
    x = np.arange(len(y))
    selection = np.logical_not(np.isnan(y))
    x = x[selection]
    y = y[selection]
    return x, y


def butterworth(X, sr, cut_offs, order=-1, filter_type='low'):
    """Apply butterworth IIR filter to a 2D array

    Args:
        X (numpy.ndarray): A 1D or 2D numpy array
        sr (float): sampling rate in Hz
        cut_offs (float or list): cut off frequencies in Hz
        order (float): order of the filter

    Keyword Arguments:
        filter_type (str): type of the filter. Could be 'low', 'high', 'pass' or 'stop'

    Returns:
        filtered_X (numpy.ndarray): A numpy array with the same shape as X
    """
    Wn = _get_wn(cut_offs, sr)
    B, A = sp_signal.butter(order, Wn, btype=filter_type, output='ba')
    result = sp_signal.filtfilt(B, A, X, axis=0, method='pad', padtype='even')
    return result


def resample(X, sr, new_sr):
    """Apply resampling IIR filter to a 2D array

    Args:
        X (numpy.ndarray): A 1D or 2D numpy array
        sr (float): sampling rate in Hz
        new_sr (float): resampled sampling rate in Hz

    Returns:
        resampled_X (numpy.ndarray): A numpy array with the new sampling rate
    """
    lcm = np.lcm(sr, new_sr)
    up = lcm / sr
    down = lcm / new_sr
    # TODO: use prime factorization to generate a cascade of downsampling filter
    new_X = sp_signal.resample_poly(X, up=up, down=down, axis=0)
    return new_X


def resample_timestamps(ts, new_n):
    """Resample timestamps to new sampling rate

    This function should be used together with `arus.extensions.numpy.resample`.

    Args:
        ts (numpy.ndarray): A 1D numpy array storing timestamps in `np.datetime64` format
        new_n (int): number of samples in the resampled sequence

    Returns:
        resampled_ts (numpy.ndarray): A 1D numpy array with the resampled timestamps
    """
    assert type(ts[0]) == np.datetime64
    st = ts[0].astype('datetime64[ms]').astype('float64')
    et = ts[-1].astype('datetime64[ms]').astype('float64')
    new_ts = np.linspace(st, et, num=new_n)
    vf = np.vectorize(lambda x: np.datetime64(int(x), 'ms'))
    new_ts = vf(new_ts)
    return new_ts


def _get_wn(cut_offs, sr):
    nyquist = sr / 2.0
    return atleast_scalar(cut_offs) / nyquist


def detect_peaks(x, mph=None, mpd=1, threshold=0, edge='rising',
                 kpsh=False, valley=False, show=False, ax=None):
    """Detect peaks in data based on their amplitude and other features.

    Parameters
    ----------
    x : 1D array_like
        data.
    mph : {None, number}, optional (default = None)
        detect peaks that are greater than minimum peak height.
    mpd : positive integer, optional (default = 1)
        detect peaks that are at least separated by minimum peak distance (in
        number of data).
    threshold : positive number, optional (default = 0)
        detect peaks (valleys) that are greater (smaller) than `threshold`
        in relation to their immediate neighbors.
    edge : {None, 'rising', 'falling', 'both'}, optional (default = 'rising')
        for a flat peak, keep only the rising edge ('rising'), only the
        falling edge ('falling'), both edges ('both'), or don't detect a
        flat peak (None).
    kpsh : bool, optional (default = False)
        keep peaks with same height even if they are closer than `mpd`.
    valley : bool, optional (default = False)
        if True (1), detect valleys (local minima) instead of peaks.
    show : bool, optional (default = False)
        if True (1), plot data in matplotlib figure.
    ax : a matplotlib.axes.Axes instance, optional (default = None).

    Returns
    -------
    ind : 1D array_like
        indeces of the peaks in `x`.

    Notes
    -----
    The detection of valleys instead of peaks is performed internally by simply
    negating the data: `ind_valleys = detect_peaks(-x)`

    The function can handle NaN's 

    See this IPython Notebook [1]_.

    References
    ----------
    .. [1] http://nbviewer.ipython.org/github/demotu/BMC/blob/master/notebooks/DetectPeaks.ipynb

    Examples
    --------
    >>> from detect_peaks import detect_peaks
    >>> x = np.random.randn(100)
    >>> x[60:81] = np.nan
    >>> # detect all peaks and plot data
    >>> ind = detect_peaks(x, show=True)
    >>> print(ind)

    >>> x = np.sin(2*np.pi*5*np.linspace(0, 1, 200)) + np.random.randn(200)/5
    >>> # set minimum peak height = 0 and minimum peak distance = 20
    >>> detect_peaks(x, mph=0, mpd=20, show=True)

    >>> x = [0, 1, 0, 2, 0, 3, 0, 2, 0, 1, 0]
    >>> # set minimum peak distance = 2
    >>> detect_peaks(x, mpd=2, show=True)

    >>> x = np.sin(2*np.pi*5*np.linspace(0, 1, 200)) + np.random.randn(200)/5
    >>> # detection of valleys instead of peaks
    >>> detect_peaks(x, mph=0, mpd=20, valley=True, show=True)

    >>> x = [0, 1, 1, 0, 1, 1, 0]
    >>> # detect both edges
    >>> detect_peaks(x, edge='both', show=True)

    >>> x = [-2, 1, -2, 2, 1, 1, 3, 0]
    >>> # set threshold = 2
    >>> detect_peaks(x, threshold = 2, show=True)
    """
    x = np.atleast_1d(x).astype('float64')
    if x.size < 3:
        return np.array([], dtype=int)
    if valley:
        x = -x
    # find indexes of all peaks
    dx = x[1:] - x[:-1]
    # handle NaN's
    indnan = np.where(np.isnan(x))[0]
    if indnan.size:
        x[indnan] = np.inf
        dx[np.where(np.isnan(dx))[0]] = np.inf
    ine, ire, ife = np.array([[], [], []], dtype=int)
    if not edge:
        ine = np.where((np.hstack((dx, 0)) < 0) & (np.hstack((0, dx)) > 0))[0]
    else:
        if edge.lower() in ['rising', 'both']:
            ire = np.where((np.hstack((dx, 0)) <= 0) &
                           (np.hstack((0, dx)) > 0))[0]
        if edge.lower() in ['falling', 'both']:
            ife = np.where((np.hstack((dx, 0)) < 0) &
                           (np.hstack((0, dx)) >= 0))[0]
    ind = np.unique(np.hstack((ine, ire, ife)))
    # handle NaN's
    if ind.size and indnan.size:
        # NaN's and values close to NaN's cannot be peaks
        ind = ind[np.in1d(ind, np.unique(
            np.hstack((indnan, indnan-1, indnan+1))), invert=True)]
    # first and last values of x cannot be peaks
    if ind.size and ind[0] == 0:
        ind = ind[1:]
    if ind.size and ind[-1] == x.size-1:
        ind = ind[:-1]
    # remove peaks < minimum peak height
    if ind.size and mph is not None:
        ind = ind[x[ind] >= mph]
    # remove peaks - neighbors < threshold
    if ind.size and threshold > 0:
        dx = np.min(np.vstack([x[ind]-x[ind-1], x[ind]-x[ind+1]]), axis=0)
        ind = np.delete(ind, np.where(dx < threshold)[0])
    # detect small peaks closer than minimum peak distance
    if ind.size and mpd > 1:
        ind = ind[np.argsort(x[ind])][::-1]  # sort ind by peak height
        idel = np.zeros(ind.size, dtype=bool)
        for i in range(ind.size):
            if not idel[i]:
                # keep peaks with the same height if kpsh is True
                idel = idel | (ind >= ind[i] - mpd) & (ind <= ind[i] + mpd) \
                    & (x[ind[i]] > x[ind] if kpsh else True)
                idel[i] = 0  # Keep current peak
        # remove the small peaks and sort back the indexes by their occurrence
        ind = np.sort(ind[~idel])

    if show:
        if indnan.size:
            x[indnan] = np.nan
        if valley:
            x = -x
        _plot_peaks(x, mph, mpd, threshold, edge, valley, ax, ind)
    return ind


def _plot_peaks(x, mph, mpd, threshold, edge, valley, ax, ind):
    """Plot results of the detect_peaks function, see its help."""
    try:
        import matplotlib.pyplot as plt
    except ImportError:
        print('matplotlib is not available.')
    else:
        if ax is None:
            _, ax = plt.subplots(1, 1, figsize=(8, 4))

        ax.plot(x, 'b', lw=1)
        if ind.size:
            label = 'valley' if valley else 'peak'
            label = label + 's' if ind.size > 1 else label
            ax.plot(ind, x[ind], '+', mfc=None, mec='r', mew=2, ms=8,
                    label='%d %s' % (ind.size, label))
            ax.legend(loc='best', framealpha=.5, numpoints=1)
        ax.set_xlim(-.02*x.size, x.size*1.02-1)
        ymin, ymax = x[np.isfinite(x)].min(), x[np.isfinite(x)].max()
        yrange = ymax - ymin if ymax > ymin else 1
        ax.set_ylim(ymin - 0.1*yrange, ymax + 0.1*yrange)
        ax.set_xlabel('Data #', fontsize=14)
        ax.set_ylabel('Amplitude', fontsize=14)
        mode = 'Valley detection' if valley else 'Peak detection'
        ax.set_title("%s (mph=%s, mpd=%d, threshold=%s, edge='%s')"
                     % (mode, str(mph), mpd, str(threshold), edge))
        plt.show()


def apply_over_subwins(X, func, subwins=None, subwin_samples=None, has_names=False, **kwargs):
    X = atleast_float_2d(X)
    if subwins is not None:
        # compute the length of each sub window if the number of sub windows is given.
        if subwins == 0:
            # if subwins is zero, treat as one
            subwins = 1
            win_length = X.shape[0]
        else:
            win_length = int(np.floor(X.shape[0] / subwins))
            if win_length == 0:
                # if subwins is zero, treat as one
                subwins = 1
                win_length = X.shape[0]
    elif subwin_samples is not None:
        # or if the number of samples of each sub window is provided, compute the number of sub windows.
        if subwin_samples == 0:
            subwins = 1
            win_length = X.shape[0]
        else:
            win_length = subwin_samples
            subwins = int(np.floor(X.shape[0] / subwin_samples))
            if subwins == 0:
                # if subwins is zero, treat as one
                subwins = 1
                win_length = X.shape[0]
    else:
        # or treat the entire input array as a single sub window
        subwins = 1
        win_length = X.shape[0]

    # Use the sub windows evenly reside in the middle of the bigger window
    start_index = np.ceil((X.shape[0] - subwins * win_length) / 2)

    result = []
    for i in range(subwins):
        indices = int(start_index) + np.array(range(
            i * win_length,
            (i + 1) * win_length
        ))
        subwin_X = X[indices, :]
        if has_names:
            subwin_result, func_names = func(subwin_X, **kwargs)
        else:
            subwin_result = func(subwin_X, **kwargs)
        subwin_result = atleast_float_2d(subwin_result)
        result.append(subwin_result)
    # each row is the result from one sub window
    final_result = np.concatenate(result, axis=0)
    if has_names:
        return final_result, func_names
    else:
        return final_result


def vector_magnitude(X):
    X = atleast_float_2d(X)
    result = la.norm(X, ord=2, axis=1, keepdims=True)
    return result


def flip_and_swap(X, x_flip, y_flip, z_flip):
    X = atleast_float_2d(X)
    X_clone = np.copy(X)
    x = np.copy(X_clone[:, 0])
    y = np.copy(X_clone[:, 1])
    z = np.copy(X_clone[:, 2])
    x_flip = x_flip.lower()
    y_flip = y_flip.lower()
    z_flip = z_flip.lower()
    if x_flip == 'x':
        X_clone[:, 0] = x
    elif x_flip == '-x':
        X_clone[:, 0] = -x
    elif x_flip == 'y':
        X_clone[:, 0] = y
    elif x_flip == '-y':
        X_clone[:, 0] = -y
    elif x_flip == 'z':
        X_clone[:, 0] = z
    elif x_flip == '-z':
        X_clone[:, 0] = -z

    if y_flip == 'x':
        X_clone[:, 1] = x
    elif y_flip == '-x':
        X_clone[:, 1] = -x
    elif y_flip == 'y':
        X_clone[:, 1] = y
    elif y_flip == '-y':
        X_clone[:, 1] = -y
    elif y_flip == 'z':
        X_clone[:, 1] = z
    elif y_flip == '-z':
        X_clone[:, 1] = -z

    if z_flip == 'x':
        X_clone[:, 2] = x
    elif z_flip == '-x':
        X_clone[:, 2] = -x
    elif z_flip == 'y':
        X_clone[:, 2] = y
    elif z_flip == '-y':
        X_clone[:, 2] = -y
    elif z_flip == 'z':
        X_clone[:, 2] = z
    elif z_flip == '-z':
        X_clone[:, 2] = -z

    return X_clone




from .. import generator
from .. import moment
import numpy as np
import pandas as pd
import time


def test_mHealthSensorFileGenerator(spades_lab_data):
    sensor_files = [spades_lab_data['subjects']
                    ['SPADES_1']['sensors']['DW'][0]]
    sizes = []
    gen = generator.MhealthSensorFileGenerator(
        *sensor_files, buffer_size=18000)
    gen.run()
    for data, _ in gen.get_result():
        if data is None:
            break
        assert type(data) == pd.DataFrame
        sizes.append(data.shape[0])
    sizes = sizes[:-1]
    assert np.all(np.array(sizes) == 18000)

    sizes = []
    gen = generator.MhealthSensorFileGenerator(
        *sensor_files, buffer_size=18000)
    gen.run()
    while True:
        try:
            data, _ = next(gen.get_result())
            if data is None:
                break
            assert type(data) == pd.DataFrame
            sizes.append(data.shape[0])
        except StopIteration:
            break
    sizes = sizes[:-1]
    assert np.all(np.array(sizes) == 18000)


def test_mHealthAnnotationFileGenerator(spades_lab_data):
    annotation_files = spades_lab_data['subjects']['SPADES_1']['annotations']['SPADESInLab']
    sizes = []
    gen = generator.MhealthAnnotationFileGenerator(
        *annotation_files, buffer_size=5)
    gen.run()
    for data, _ in gen.get_result():
        if data is None:
            break
        assert type(data) == pd.DataFrame
        sizes.append(data.shape[0])
    sizes = sizes[:-1]
    assert np.all(np.array(sizes) == 5)

    sizes = []
    gen = generator.MhealthAnnotationFileGenerator(
        *annotation_files, buffer_size=5)
    gen.run()
    while True:
        try:
            data, _ = next(gen.get_result())
            if data is None:
                break
            assert type(data) == pd.DataFrame
            sizes.append(data.shape[0])
        except StopIteration:
            break
    sizes = sizes[:-1]
    assert np.all(np.array(sizes) == 5)


def test_RandomAccelDataGenerator():
    # default setting
    sr = 3600
    grange = 4
    start_time = None
    buffer_size = int(3600 / 2)
    sigma = 1
    max_samples = buffer_size * 2
    gen = generator.RandomAccelDataGenerator(
        sr=sr,
        grange=grange,
        st=start_time,
        buffer_size=buffer_size,
        sigma=sigma,
        max_samples=max_samples)
    gen.run()
    for data, _ in gen.get_result():
        if data is None:
            break
        mean_data = np.mean(data.values[:, 1:], axis=0)
        std_data = np.std(data.iloc[:, 1:].values, axis=0)
        duration = (data.iloc[-1, 0] - data.iloc[0, 0]) / \
            pd.Timedelta(1, unit='seconds')
        np.testing.assert_array_almost_equal(
            mean_data, np.array([0, 0, 0]), decimal=1)
        np.testing.assert_array_almost_equal(
            std_data, np.array([sigma, sigma, sigma]),  decimal=1)
        np.testing.assert_almost_equal(
            duration, buffer_size / sr, decimal=1)


def test_RandomAnnotationDataGenerator():
    # default setting
    duration_mu = 5
    duration_sigma = 1
    start_time = None
    num_mu = 3
    labels = ['Sitting', 'Standing', 'Lying']
    max_samples = 50

    durations = []
    rows = []
    gen = generator.RandomAnnotationDataGenerator(labels=labels,
                                                  duration_mu=duration_mu, duration_sigma=duration_sigma, st=start_time, num_mu=num_mu, max_samples=max_samples)
    gen.run()
    for data, _ in gen.get_result():
        if data is None:
            break
        durations += moment.Moment.get_durations(
            data['START_TIME'], data['STOP_TIME'], unit='s')
        rows.append(data.shape[0])
    duration_mean = np.mean(durations)
    rows_mean = np.mean(rows)
    np.testing.assert_almost_equal(duration_mean, duration_mu, decimal=0)
    np.testing.assert_almost_equal(rows_mean, num_mu, decimal=0)




from .. import stream2
from .. import generator
from .. import segmentor
from .. import synchronizer
from .. import processor
from .. import scheduler
from .. import pipeline
from .. import node
from .. import mh
from .. import developer
import pytest
import pandas as pd
import numpy as np


@pytest.fixture
def sample_pipeline(spades_lab_data):
    def compute_mean(values_list, src=None, **context):
        import numpy as np
        import pandas as pd
        results = []
        for values, data_id in zip(values_list, context['data_ids']):
            data = values.values[:, 1:]
            col_names = [data_id + '_' +
                         name for name in ['MEAN_0', 'MEAN_1', 'MEAN_2']]
            result = pd.DataFrame(data=[np.mean(
                data, axis=0)], columns=col_names)
            results.append(result)

        result = pd.concat(results, axis=1)
        result.insert(0, 'START_TIME', context['start_time'])
        result.insert(1, 'STOP_TIME', context['stop_time'])
        return result, context
    dw_data = spades_lab_data['subjects']['SPADES_2']['sensors']['DW']
    da_data = spades_lab_data['subjects']['SPADES_2']['sensors']['DA']

    start_time = mh.get_session_start_time(
        'SPADES_2', spades_lab_data['meta']['root'], round_to='minute')

    dw_stream = stream2.Stream(
        generator.MhealthSensorFileGenerator(*dw_data, buffer_size=18000), segmentor.SlidingWindowSegmentor(window_size=2), name='dw-stream')
    dw_stream.set_essential_context(start_time=start_time, stream_id='DW')

    da_stream = stream2.Stream(
        generator.MhealthSensorFileGenerator(*da_data, buffer_size=18000), segmentor.SlidingWindowSegmentor(window_size=2), name='dw-stream')
    da_stream.set_essential_context(start_time=start_time, stream_id='DA')

    sync = synchronizer.Synchronizer()
    sync.add_sources(n=2)
    proc = processor.Processor(
        compute_mean, mode=scheduler.Scheduler.Mode.PROCESS, scheme=scheduler.Scheduler.Scheme.SUBMIT_ORDER, max_workers=10)

    pip = node.Node(op=pipeline.Pipeline(dw_stream, da_stream, synchronizer=sync,
                                         processor=proc, name='test-pipeline'),
                    t=node.Node.Type.INPUT, name='test-pipeline')
    return pip


def test_pipeline_lifecycle(sample_pipeline):
    sample_pipeline.start()
    results = []
    while True:
        pack = next(sample_pipeline.produce())
        if pack.signal == node.Node.Signal.DATA:
            if pack.values is not None:
                results.append(pack.values)
            if len(results) == 3 or pack.values is None:
                break
    sample_pipeline.stop()
    result = pd.concat(results, axis=0)
    np.testing.assert_array_equal(result.shape, [3, 8])
    sample_pipeline.get_op().shutdown()



"""
Module that provides environment variable and path settings for arus package

Author: Qu Tang
Date: 01/28/2020
License: GNU v3
"""


import os

_ARUS_DATA_ = 'ARUS_DATA'
_CACHE_ = 'CACHE'


def get_data_home(data_home=None):
    """Return the path of the seaborn data directory.
    This is used by the ``load_dataset`` function.
    If the ``data_home`` argument is not specified, the default location
    is ``~/arus-data``.
    Alternatively, a different default location can be specified using the
    environment variable ``ARUS_DATA``.
    """
    if data_home is None:
        data_home = os.environ.get(_ARUS_DATA_,
                                   os.path.join('~', 'arus-data'))
    data_home = os.path.expanduser(data_home)
    if not os.path.exists(data_home):
        os.makedirs(data_home)
    return data_home


def get_cache_home(cache_home=None):
    if cache_home is None:
        cache_home = os.environ.get(_CACHE_,
                                    os.path.join('~', 'arus-cache'))
    cache_home = os.path.expanduser(cache_home)
    if not os.path.exists(cache_home):
        os.makedirs(cache_home)
    return cache_home



"""
Demonstration of the usage of arus.core.Pipeline (2)
====================================================

The pipeline uses one annotation stream and one sensor stream as input.
"""

from arus.core.pipeline import Pipeline
import arus
import datetime as dt
import pandas as pd
import numpy as np


def _master_pipeline_processor(chunk_list, **kwargs):
    import pandas as pd
    import numpy as np
    result = {'HEADER_TIME_STAMP': [], 'START_TIME': [],
              'STOP_TIME': [], 'VALUE': [], 'ANNOTATION': []}
    for data, st, et, prev_st, prev_et, name in chunk_list:
        if len(result['START_TIME']) == 0 or result['START_TIME'][-1] != st:
            result['HEADER_TIME_STAMP'].append(st)
            result['START_TIME'].append(st)
            result['STOP_TIME'].append(et)
        if name == 'annotation-stream':
            result['ANNOTATION'].append('-'.join(data.iloc[:, 3].values))
        elif name == 'feature-pipeline':
            result['VALUE'].append(data.iloc[0, 3])
    result = pd.DataFrame.from_dict(result)

    return result


def _feature_pipeline_processor(chunk_list, **kwargs):
    import pandas as pd
    import numpy as np
    import arus
    result = {'HEADER_TIME_STAMP': [],
              'START_TIME': [], 'STOP_TIME': [], 'VALUE': []}
    for data, st, et, prev_st, prev_et, name in chunk_list:
        result['HEADER_TIME_STAMP'].append(st)
        result['START_TIME'].append(st)
        result['STOP_TIME'].append(et)
        vm_values = arus.ext.numpy.vector_magnitude(data.iloc[:, 1:].values)
        values, name = arus.accel.mean(vm_values)
        result['VALUE'].append(values[0, 0])
    result = pd.DataFrame.from_dict(result)
    return result


if __name__ == "__main__":
    # test on one annotation stream and one sensor stream + feature pipeline

    window_size = 12.8
    sr = 80
    start_time = dt.datetime.now()
    gr1 = arus.generator.RandomAccelDataGenerator(
        sr=sr, grange=8, sigma=1, buffer_size=100)
    seg1 = arus.segmentor.SlidingWindowSegmentor(window_size)
    stream1 = arus.Stream(gr1, seg1, name='sensor-stream')

    gr2 = arus.generator.RandomAnnotationDataGenerator(
        labels=["Sitting", 'Standing', 'Lying', 'Walking', 'Running'], duration_mu=8, duration_sigma=2, num_mu=3, buffer_size=100)
    seg2 = arus.segmentor.SlidingWindowSegmentor(
        window_size, st_col=1, et_col=2)
    stream2 = arus.Stream(gr2, seg2, name='annotation-stream')

    feat_pipeline = Pipeline(
        max_processes=2, scheduler='processes', name='feature-pipeline')

    feat_pipeline.add_stream(stream1)
    feat_pipeline.set_processor(_feature_pipeline_processor)
    master_pipeline = Pipeline(
        max_processes=2, scheduler='processes', name='master-pipeline')
    master_pipeline.add_stream(stream2)
    master_pipeline.add_stream(feat_pipeline)
    master_pipeline.set_processor(_master_pipeline_processor)
    master_pipeline.start(start_time=start_time)
    results = []
    for result, st, et, prev_st, prev_et, name in master_pipeline.get_iterator():
        results.append(result)
        if len(results) == 10:
            break
        print(len(results))
    print(master_pipeline.stop())
    print(pd.concat(results, axis=0, sort=False))



"""
Computing features about different versions of counts
"""
import numpy as np
from .. import extensions as ext

COUNT_FEATURE_NAME_PREFIX = [
    'ENMO'
]


def count_features(X, selected=COUNT_FEATURE_NAME_PREFIX):
    count_funcs = [
        enmo
    ]

    fv = []
    fv_names = []
    for func, prefix in zip(count_funcs, COUNT_FEATURE_NAME_PREFIX):
        if prefix in selected:
            result, names = func(X)
            fv.append(result)
            fv_names += names

    if len(fv) == 0:
        return None, None

    result = np.concatenate(fv, axis=1)
    return result, fv_names


def enmo(X):
    """
    Computing ENMO value of accelerometer data

    Arguments:
        X {numpy.ndarray} -- 2D numpy array M x N, M is the number of samples and N is the dimension of the data.

    Returns:
        [numpy.ndarray] -- 1 x 1
    """
    X = ext.numpy.atleast_float_2d(X)
    vm = ext.numpy.vector_magnitude(X)
    result = np.nanmean(np.clip(vm - 1, a_min=0, a_max=None),
                        axis=0, keepdims=True)
    return result, f'{COUNT_FEATURE_NAME_PREFIX[0]}_0'



import queue
from loguru import logger
import threading
import time
import enum
import collections


class Node:
    """The elementary node to build up computational graph/pipeline
    """

    Pack = collections.namedtuple(
        "Package", "values signal context src")

    class Status(enum.Enum):
        """Operator Status codes."""
        OFF = enum.auto()
        ON = enum.auto()
        START = enum.auto()
        RUN = enum.auto()
        STOP = enum.auto()

    class Signal(enum.Enum):
        "Operator Signal codes."
        STOP = enum.auto()
        WAIT = enum.auto()
        DATA = enum.auto()

    class Type(enum.Enum):
        INPUT = enum.auto()
        OUTPUT = enum.auto()
        PIPE = enum.auto()

    def __init__(self, op, t=Type.PIPE, name='default'):
        self._name = name
        self._input_buffer = queue.Queue()
        self._output_buffer = queue.Queue()
        self._operator = op
        self._produce_thread = None
        self._result_thread = None
        self._status = Node.Status.ON
        self._type = t

    def get_name(self):
        return self._name

    def get_type(self):
        return self._type

    def toggle(self, status=Status.OFF):
        self._status = Node.Status.ON

    def get_op(self):
        return self._operator

    def start(self):
        if self._status == Node.Status.OFF:
            logger.warning('Please turn on the operator at first')
            return
        self._status = Node.Status.START
        logger.info(f'{self._name} Operator is starting.')
        self._produce_thread = threading.Thread(
            target=self._produce, name=self._name + '-produce')
        self._produce_thread.daemon = True
        self._produce_thread.start()
        self._result_thread = threading.Thread(
            target=self._get_result, name=self._name + '-result'
        )
        self._result_thread.daemon = True
        self._result_thread.start()
        while not self._produce_thread.is_alive() or not self._result_thread.is_alive():
            time.sleep(0.1)
        logger.info(f'{self._name} Operator started.')

    def stop(self):
        logger.info(f'{self._name} Operator is stopping.')
        self._operator.stop()
        self._status = Node.Status.STOP
        self._produce_thread.join()
        self._result_thread.join()
        logger.info(f'{self._name} Operator thread stopped.')
        with self._input_buffer.mutex:
            self._input_buffer.queue.clear()
        with self._output_buffer.mutex:
            self._output_buffer.queue.clear()
        self._status = Node.Status.ON
        logger.info(f'{self._name} Operator stopped.')

    def consume(self, pack):
        if self._type != Node.Type.INPUT:
            if pack.signal in [Node.Signal.WAIT, Node.Signal.STOP]:
                pass
            else:
                self._input_buffer.put(pack)
        else:
            logger.warning('INPUT operator does not support consume method')

    def produce(self):
        """A python generator function to get the output of the operator.

        Returns:
            output of the operator.
        """
        try:
            data = self._output_buffer.get(timeout=0.1)
            yield data
        except queue.Empty:
            yield Node.Pack(values=None, signal=Node.Signal.WAIT, context={}, src=self._name)
        finally:
            if self._status == Node.Status.STOP:
                return
            pass

    def get_status(self) -> "Node.Status":
        """Get the status code of the operator.

        Returns:
            The status code of the operator.
        """
        return self._status

    def _get_result(self):
        while True:
            for values, context in self._operator.get_result():
                if self._status == Node.Status.STOP:
                    break
                else:
                    self._output_buffer.put(
                        Node.Pack(values=values, signal=Node.Signal.DATA,
                                  context=context, src=self._name)
                    )
            if self._status == Node.Status.STOP:
                break
            else:
                pass
        self._output_buffer.put(
            Node.Pack(values=None, signal=Node.Signal.STOP, context={}, src=self._name))

    def _produce_from_input(self):
        self._operator.run()

    def _produce_from_pipe(self):
        src = self._name
        while True:
            try:
                data = self._input_buffer.get(timeout=0.1)
                self._operator.run(
                    data.values, context=data.context, src=data.src)
            except queue.Empty:
                pass
            finally:
                if self._status == Node.Status.STOP:
                    break
        self._output_buffer.put(
            Node.Pack(values=None, signal=Node.Signal.STOP, context={}, src=src))

    def _produce_from_output(self):
        while True:
            try:
                data = self._input_buffer.get(timeout=0.1)
                self._operator.run(
                    data.values, context=data.context, src=data.src)
            except queue.Empty:
                pass
            finally:
                if self._status == Node.Status.STOP:
                    break
            self._output_buffer.put(
                Node.Pack(values=None, signal=Node.Signal.STOP, context={}, src=self._name))

    def _produce(self):
        logger.info(f'{self._name} Operator thread started.')
        if self._type == Node.Type.INPUT:
            self._produce_from_input()
        elif self._type == Node.Type.OUTPUT:
            self._produce_from_output()
        elif self._type == Node.Type.PIPE:
            self._produce_from_pipe()
        else:
            logger.error(f'Unknown type for the operator {self._name}')
        logger.info(f'{self._name} Operator thread is stopping.')




from .. import stream2
from .. import generator
from .. import segmentor
from .. import developer
from .. import node
import pytest


@pytest.fixture(params=['infinite', 'finite'])
def test_stream(request):
    if request.param == 'infinite':
        gr = generator.RandomAccelDataGenerator(50, buffer_size=50)
    else:
        gr = generator.RandomAccelDataGenerator(
            50, buffer_size=50, max_samples=300)
    seg = segmentor.SlidingWindowSegmentor(window_size=1)
    stream_op = stream2.Stream(gr, seg, name='{}-stream'.format(request.param))
    stream_op.set_context(ref_start_time=None, data_id='DW')
    stream = node.Node(
        op=stream_op,
        t=node.Node.Type.INPUT, name='{}-stream'.format(request.param))
    return stream


def test_stream_lifecycle(test_stream):
    test_stream.start()
    results = []
    while True:
        pack = next(test_stream.produce())
        if pack.signal == node.Node.Signal.DATA:
            if pack.values is not None:
                assert 'start_time' in pack.context
                assert 'data_id' in pack.context
                results.append(pack)
            if len(results) == 10 or pack.values is None:
                break
    if 'infinite' in test_stream._name:
        assert len(results) == 10
    else:
        assert len(results) == 6
    test_stream.stop()



"""
scheduler class that handles the execution order of functions when using multiprocessing.

Author: Qu Tang

Date: 02/07/2020

License: GNU v3
"""
import loky
import enum
from concurrent import futures
import queue
import threading
from loguru import logger
from dataclasses import dataclass, field
from typing import Any
import tqdm


class Scheduler:
    """Scheduler to support different scheduling schemes when running tasks on subprocess workers.
    """

    @dataclass(order=True)
    class PrioritizedItem:
        priority: int
        item: Any = field(compare=False)

    class ClosedError(Exception):
        pass

    class ResultNotAvailableError(Exception):
        pass

    class Mode(enum.Enum):
        THREAD = enum.auto()
        PROCESS = enum.auto()

    class Scheme(enum.Enum):
        AFTER_PREVIOUS_DONE = enum.auto()
        EXECUTION_ORDER = enum.auto()
        SUBMIT_ORDER = enum.auto()

    def __init__(self, mode: "Scheduler.Mode" = Mode.PROCESS, scheme: "Scheduler.Scheme" = Scheme.EXECUTION_ORDER, max_workers: int = None):
        """Create Scheduler instance.

        Arguments:
            mode: run tasks in process or in thread.
            scheme: the scheduling scheme when executing tasks.
            max_workers: the max number of workers. If None, it will be cpu count - 4 workers.
        """

        if max_workers is None:
            max_workers = loky.cpu_count() - 4

        if mode == Scheduler.Mode.PROCESS:
            self._executor = loky.get_reusable_executor(
                max_workers=max_workers, timeout=10)
            self._module = loky
        elif mode == Scheduler.Mode.THREAD:
            self._executor = futures.ThreadPoolExecutor(
                max_workers=max_workers, thread_name_prefix='default-scheduler')
            self._module = futures
        self._tasks = queue.Queue()
        self._scheme = scheme
        self._close = False
        self._results = queue.PriorityQueue()
        self._priority_number = 0

    def reset(self):
        """Reset scheduler.

        Clear results and tasks queue.
        """
        with self._results.mutex:
            self._results.queue.clear()
        with self._tasks.mutex:
            self._tasks.queue.clear()
        self._close = False
        self._priority_number = 0

    def close(self):
        """Close a scheduler from accepting new tasks.
        """
        self._close = True

    def shutdown(self):
        """Shut down a scheduler.

        Reset the scheduler and shut down the inner executor.
        """
        self.reset()
        self._executor.shutdown(wait=True)

    def submit(self, func: object, *args: object, **kwargs: object) -> futures.Future:
        """Submit a new task to the scheduler.

        Arguments:
            func: the task function to be submitted to the scheduler. It should be picklable. It should return a tuple including two items. The first item is the computed values and the second is the context as a dict to be passed with the computed values.
            args: the positional arguments passed to `func`.
            kwargs: the keyword arguments passed to `func`.

        Raises:
            Scheduler.ClosedError: Raise when trying to submit task to a closed scheduler.

        Returns:
            A future instance of the pending submitted task.
        """
        if self._close:
            logger.warning('Scheduler is closed for new tasks.')
            raise Scheduler.ClosedError('Scheduler is closed for new tasks.')
        if self._scheme == Scheduler.Scheme.AFTER_PREVIOUS_DONE:
            if not self._tasks.empty():
                try:
                    prev_task = self._tasks.get()
                    self._add_to_results(prev_task)
                except queue.Empty:
                    pass
            task = self._executor.submit(func, *args, **kwargs)
            self._tasks.put(task)
        else:
            task = self._executor.submit(func, *args, **kwargs)
            self._tasks.put(task)
            task.add_done_callback(self._add_to_results)
        return task

    def _add_to_results(self, task):
        if self._scheme == Scheduler.Scheme.EXECUTION_ORDER:
            item = Scheduler.PrioritizedItem(
                priority=self._priority_number, item=task)
            self._results.put(item)
            self._priority_number += 1
        elif self._scheme == Scheduler.Scheme.SUBMIT_ORDER:
            i = self._tasks.queue.index(task)
            item = Scheduler.PrioritizedItem(priority=i, item=task)
            self._results.put(item)
        elif self._scheme == Scheduler.Scheme.AFTER_PREVIOUS_DONE:
            task.result()
            item = Scheduler.PrioritizedItem(
                priority=self._priority_number, item=task)
            self._results.put(item)
            self._priority_number += 1

    def get_all_remaining_results(self) -> list:
        """Get all remaining results from the submitted tasks.

        Returns:
            A list of results of the remaining submitted tasks.
        """
        self._close = True
        self._module.wait(self._tasks.queue)
        results = []
        with tqdm.tqdm(total=self._tasks.qsize()) as bar:
            if self._scheme == Scheduler.Scheme.AFTER_PREVIOUS_DONE:
                while not self._results.empty():
                    bar.set_description(f'Finished tasks: {len(results)}')
                    t = self._results.get()
                    results.append(t.item.result())
                    bar.update()
                t = self._tasks.get()
                results.append(t.result())
            else:
                while not self._results.empty():
                    bar.set_description(f'Finished tasks: {len(results)}')
                    t = self._results.get()
                    self._tasks.get()
                    results.append(t.item.result())
                    bar.update()
        return results

    def get_result(self, timeout: float = None) -> tuple:
        """Get the next result of the submitted tasks.

        Arguments:
            timeout: the time out in seconds to wait for the next results. If it is `None`, wait in infinite time.

        Raises:
            Scheduler.ResultNotAvailableError: Raise when result is not available yet.

        Returns:
            The next result of the submitted tasks. It should be a tuple with two items. The first is the computed values and the second is the context dict.
        """
        try:
            if self._scheme == Scheduler.Scheme.EXECUTION_ORDER:
                task = self._results.get(timeout=timeout)
                self._tasks.get()
                return task.item.result()
            elif self._scheme == Scheduler.Scheme.SUBMIT_ORDER:
                if not self._results.empty():
                    first_task = self._results.queue[0]
                    if self._tasks.queue.index(first_task.item) != 0:
                        raise queue.Empty
                    else:
                        first_task = self._results.get()
                        self._tasks.get()
                        result = first_task.item.result()
                        return result
                else:
                    raise queue.Empty
            elif self._scheme == Scheduler.Scheme.AFTER_PREVIOUS_DONE:
                if not self._results.empty():
                    task = self._results.get()
                    return task.item.result()
                elif not self._tasks.empty():
                    return self._tasks.get().result()
                else:
                    raise queue.Empty
        except queue.Empty:
            raise Scheduler.ResultNotAvailableError(
                'None of the results are available.')



"""
Demonstration of the usage of MetaWearScanner
====================================================
"""
from arus.plugins.metawear import MetaWearScanner


if __name__ == "__main__":
    scanner = MetaWearScanner()
    print(scanner.get_nearby_devices(max_devices=10))



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['arus',
 'arus.accelerometer',
 'arus.core',
 'arus.core.tests',
 'arus.dataset',
 'arus.extensions',
 'arus.feature_vector',
 'arus.mhealth_format',
 'arus.models',
 'arus.plugins',
 'arus.spades_lab',
 'arus.tests']

package_data = \
{'': ['*'], 'arus.dataset': ['data/*'], 'arus.models': ['prebuilt/*']}

install_requires = \
['docopt>=0.6.2,<0.7.0',
 'joblib>=0.16.0,<0.17.0',
 'loguru>=0.5.0,<0.6.0',
 'loky>=2.6.0,<3.0.0',
 'matplotlib>=3.1,<4.0',
 'numpy==1.19.3',
 'pandas>=1.0.0,<2.0.0',
 'pathos>=0.2.5,<0.3.0',
 'pyarrow>=1.0.0,<2.0.0',
 'scikit-learn>=0.23.2,<0.24.0',
 'scipy>=1.3,<2.0',
 'seaborn>=0.11.0,<0.12.0',
 'tqdm>=4.48.0,<5.0.0',
 'tzlocal>=2.0.0,<3.0.0',
 'wget>=3.2,<4.0']

extras_require = \
{':python_version >= "3.6" and python_version < "3.7"': ['dataclasses>=0.7,<0.8'],
 'demo': ['playsound>=1.2.2,<2.0.0',
          'pymetawear>=0.12.0,<0.13.0',
          'pysimplegui>=4.14.1,<5.0.0'],
 'dev': ['dephell_versioning>=0.1.2,<0.2.0', 'semver>=2.10.1,<3.0.0'],
 'metawear': ['pymetawear>=0.12.0,<0.13.0'],
 'nn': ['tensorboard>=2.3.0,<3.0.0'],
 'nn:sys_platform == "linux"': ['torch>=1.6.0,<2.0.0']}

entry_points = \
{'console_scripts': ['arus = arus.cli:cli']}

setup_kwargs = {
    'name': 'arus',
    'version': '1.1.18',
    'description': 'Activity Recognition with Ubiquitous Sensing',
    'long_description': '# Overview\n\n__ARUS__ python package provides a computational and experimental framework to manage and process sensory data or wireless devices, to develop and run activity recognition algorithms on the data, and to create interactive programs using the algorithms and wireless devices.\n\nThis package is licensed under [GPL version 3](https://qutang.github.io/arus/LICENSE/).\n\n[![PyPI version](https://badge.fury.io/py/arus.svg)](https://badge.fury.io/py/arus)  \n[![Downloads](https://pepy.tech/badge/arus)](https://pepy.tech/project/arus)  \n[![deployment build](https://github.com/qutang/arus/workflows/deploy/badge.svg)](https://github.com/qutang/arus/actions?query=workflow%3Adeploy)  \n[![unittest and build test](https://github.com/qutang/arus/workflows/unittest%20and%20build%20test/badge.svg)](https://github.com/qutang/arus/actions?query=workflow%3A%22unittest+and+build+test%22)  \n[![codecov](https://codecov.io/gh/qutang/arus/branch/master/graph/badge.svg)](https://codecov.io/gh/qutang/arus)  \n\n## Prerequists\n\n```bash\npython >= 3.7.0\n```\n\n```bash\n# Need these SDKs to install arus[metawear] on Windows.\nVisual Studio C++ SDK (v14.1)\nWindows SDK (10.0.16299.0)\nWindows SDK (10.0.17763.0)\n\n# Need these packages to install arus[metawear] on Ubuntu or equivalent packages on other linux distributions.\nlibbluetooth-dev\nlibboost-all-dev\nbluez\n```\n\n## Installation\n\n```bash\n> pip install arus\n# Optionally, you may install plugins via pip extra syntax.\n> pip install arus[metawear]\n> pip install arus[demo]\n> pip install arus[dev]\n> pip install arus[nn]\n```\n\n## Optional components\n\n`arus[metawear]`: This optional component installs dependency supports for streaming data from Bluetooth metawear sensors.\n\n`arus[demo]`: This optional component installs dependency supports for running the demo app that demonstrates a real-time interactive activity recognition training and testing program.\n\n`arus[dev]`: These optional component installs dependency supports for running some package and version management functions in the `arus.dev` module.\n\n`arus[nn]`: The optional component installs dependency supports for PyTorch and Tensorboard, which are required by `arus.models.report` module. Note that for Windows, you should install the `torch` package manually using `pip` following the `pytorch.org` instruction.\n\n## Get started for development\n\n```bash\n> git clone https://github.com/qutang/arus.git\n> cd arus\n> # Install poetry python package management tool https://python-poetry.org/docs/\n> # On Linux\n> curl -sSL https://raw.githubusercontent.com/python-poetry/poetry/master/get-poetry.py | python\n> # On windows powershell\n> (Invoke-WebRequest -Uri https://raw.githubusercontent.com/python-poetry/poetry/master/get-poetry.py -UseBasicParsing).Content | python\n> # Install package dependencies\n> poetry install\n> # Install optional component dependencies\n> poetry install --extras "metawear demo dev nn"\n> # Run unit tests\n> poetry run pytest\n```\n\n### Development conventions\n\n1. Use Google\'s python coding guidance: http://google.github.io/styleguide/pyguide.html.\n2. Use `arus package release VERSION --release` to bump and tag versions. `VERSION` can be manual version code following semantic versioning, `path`, `minor`, or `major`.\n3. Changelogs are automatically generated when building the documentation website, do not create it manually.\n4. Pypi release will be handled by github action `deploy.yml`, which will be triggered whenever a new tag is pushed. Therefore, developers should only tag release versions.\n5. After commits, even not bumping and releasing the package, you may run `arus package docs --release` to update the documentation website, where the developer version changelogs will be updated immediately.',
    'author': 'qutang',
    'author_email': 'tqshelly@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/qutang/arus',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'entry_points': entry_points,
    'python_requires': '>=3.7.0,<4.0.0',
}


setup(**setup_kwargs)



"""
Computing features about accelerometer orientations
"""

import numpy as np

from .. import extensions as ext
from . import stats

ORIENTATION_FEATURE_NAME_PREFIX = [
    'MEDIAN_G_ANGLE',
    'RANGE_G_ANGLE',
    'STD_G_ANGLE'
]


def _gravity_angles(X, unit='rad'):
    X = ext.numpy.atleast_float_2d(X)
    gravity = stats.mean(X)[0]
    gravity_vm = ext.numpy.vector_magnitude(gravity)
    gravity_angles = np.arccos(
        gravity / gravity_vm) if gravity_vm != 0 else np.zeros_like(gravity)
    if unit == 'deg':
        gravity_angles = np.rad2deg(gravity_angles)
    return gravity_angles


def gravity_angles(X, subwins=None, subwin_samples=None, unit='rad'):
    result = ext.numpy.apply_over_subwins(
        X, _gravity_angles, subwins=subwins, subwin_samples=subwin_samples, unit=unit)
    final_result = np.atleast_2d(result.flatten())
    names = []
    for i in range(result.shape[0]):
        names = names + ['G_ANGLE_X_' + str(i), 'G_ANGLE_Y_' +
                         str(i), 'G_ANGLE_Z_' + str(i)]
    return final_result, names


def orientation_features(X, subwins=None, subwin_samples=None, unit='rad', selected=ORIENTATION_FEATURE_NAME_PREFIX):

    X = ext.numpy.atleast_float_2d(X)
    result = ext.numpy.apply_over_subwins(
        X, _gravity_angles, subwins=subwins, subwin_samples=subwin_samples, unit=unit)

    fv = []
    fv_names = []

    if ORIENTATION_FEATURE_NAME_PREFIX[0] in selected:
        fv.append(np.nanmedian(result, axis=0, keepdims=True))
        fv_names += [f'{ORIENTATION_FEATURE_NAME_PREFIX[0]}_{i}' for i in [0, 1, 2]]

    if ORIENTATION_FEATURE_NAME_PREFIX[1] in selected:
        fv.append(np.nanmax(
            result, axis=0, keepdims=True) - np.nanmin(result, axis=0, keepdims=True))
        fv_names += [f'{ORIENTATION_FEATURE_NAME_PREFIX[1]}_{i}' for i in [0, 1, 2]]

    if ORIENTATION_FEATURE_NAME_PREFIX[2] in selected:
        fv.append(np.nanstd(result, axis=0, keepdims=True, ddof=1))
        fv_names += [f'{ORIENTATION_FEATURE_NAME_PREFIX[2]}_{i}' for i in [0, 1, 2]]

    if len(fv) == 0:
        return None, None

    result = np.concatenate(fv, axis=1)
    return result, fv_names



from .. import generator as gr
from .. import segmentor as seg
from .. import synchronizer as sync
from .. import processor as pr
from .. import scheduler as sch
from .. import node
import numpy as np


def test_random_data_generator_operator():
    generator = gr.RandomAccelDataGenerator(sr=80, buffer_size=10)
    op = node.Node(generator, name='accel', t=node.Node.Type.INPUT)

    op.start()
    i = 0
    while True:
        data = next(op.produce())
        if data.signal == node.Node.Signal.WAIT:
            continue
        assert data.values.shape[0] == 10
        if i == 2:
            break
        i += 1
    op.stop()


def test_mhealth_file_generator_operator(spades_lab_data):
    mhealth_filepath = spades_lab_data['subjects']['SPADES_2']['sensors']['DW'][0]
    generator = gr.MhealthSensorFileGenerator(
        mhealth_filepath, buffer_size=10)

    op = node.Node(generator, name='SPADES_2-DW', t=node.Node.Type.INPUT)

    op.start()
    i = 0
    while True:
        data = next(op.produce())
        if data.signal == node.Node.Signal.WAIT:
            continue
        assert data.values.shape[0] == 10
        if i == 2:
            break
        i += 1
    op.stop()


def test_segmentor_operator(spades_lab_data):
    mhealth_filepath = spades_lab_data['subjects']['SPADES_2']['sensors']['DW'][0]
    generator = gr.MhealthSensorFileGenerator(
        mhealth_filepath, buffer_size=1800)
    generator.run()
    values, context = next(generator.get_result())
    generator.stop()
    segmentor = seg.SlidingWindowSegmentor(window_size=1)
    op = node.Node(segmentor, name='segmentor-1s')

    op.start()
    op.consume(node.Node.Pack(values=values, signal=node.Node.Signal.DATA,
                              context=context, src='SPADES_2-dw'))
    i = 0
    while True:
        data = next(op.produce())
        if data.signal == node.Node.Signal.WAIT:
            continue
        assert data.values.shape[0] == 80
        if i == 2:
            break
        i += 1
    op.stop()


def test_synchronizer_operator(spades_lab_data):
    dw_filepath = spades_lab_data['subjects']['SPADES_2']['sensors']['DW'][0]
    da_filepath = spades_lab_data['subjects']['SPADES_2']['sensors']['DA'][0]
    dw_generator = gr.MhealthSensorFileGenerator(
        dw_filepath, buffer_size=1800)
    da_generator = gr.MhealthSensorFileGenerator(
        da_filepath, buffer_size=1800)
    dw_generator.run()
    da_generator.run()
    dw_values, dw_context = next(dw_generator.get_result())
    da_values, da_context = next(da_generator.get_result())
    dw_context = {**dw_context, 'data_id': 'DW'}
    da_context = {**da_context, 'data_id': 'DA'}
    dw_generator.stop()
    da_generator.stop()

    segmentor = seg.SlidingWindowSegmentor(window_size=1)
    segmentor.run(
        dw_values, src=None, context=dw_context)
    seg_dw_values, dw_context = next(segmentor.get_result())
    segmentor = seg.SlidingWindowSegmentor(window_size=1)
    segmentor.run(
        da_values, src=None, context=da_context)
    seg_da_values, da_context = next(segmentor.get_result())

    synchronizer = sync.Synchronizer()
    synchronizer.add_sources(2)
    op = node.Node(synchronizer, name='sync-dw-da')
    op.start()
    op.consume(node.Node.Pack(values=seg_dw_values,
                              signal=node.Node.Signal.DATA, context=dw_context, src='dw'))
    op.consume(node.Node.Pack(values=seg_da_values,
                              signal=node.Node.Signal.DATA, context=da_context, src='da'))
    while True:
        data = next(op.produce())
        if data.signal == node.Node.Signal.WAIT:
            continue
        assert data.values[0].shape[0] == 80
        assert data.values[1].shape[0] == 80
        np.testing.assert_array_equal(data.context['data_ids'], ['DW', 'DA'])
        break
    op.stop()


def test_processor_operator(spades_lab_data):
    dw_filepath = spades_lab_data['subjects']['SPADES_2']['sensors']['DW'][0]
    dw_generator = gr.MhealthSensorFileGenerator(
        dw_filepath, buffer_size=1800)
    dw_generator.run()
    dw_values, dw_context = next(dw_generator.get_result())
    dw_context = {**dw_context, 'data_id': 'DW'}
    dw_generator.stop()

    def compute_mean(values, src=None, **context):
        import numpy as np
        data = values.values[:, 1:]
        result = np.mean(data, axis=0)
        return result, context

    processor = pr.Processor(compute_mean,
                             mode=sch.Scheduler.Mode.PROCESS,
                             scheme=sch.Scheduler.Scheme.SUBMIT_ORDER, max_workers=2)

    op = node.Node(processor, name='compute-mean')
    op.start()
    op.consume(node.Node.Pack(values=dw_values, signal=node.Node.Signal.DATA,
                              context=dw_context, src='dw'))
    while True:
        data = next(op.produce())
        if data.signal == node.Node.Signal.WAIT:
            continue
        assert len(data.values) == 3
        break
    op.stop()



"""
Stream from metawear device
================================================================================

"""

import arus
from datetime import datetime

if __name__ == "__main__":
    generator = arus.plugins.metawear.MetaWearAccelDataGenerator(
        "F9:DE:3A:BD:B2:84", sr=50, grange=8, max_retries=10, buffer_size=100)
    print(generator.get_device_name())
    segmentor = arus.segmentor.SlidingWindowSegmentor(window_size=4)
    stream = arus.Stream(generator, segmentor, name='metawear-stream')
    stream.start()
    i = 0
    for data, _, _, _, _, _ in stream.generate():
        if data.empty:
            print(data)
            continue
        i = i + 1
        print(data.head())
        if i == 2:
            break
    stream.stop()



"""
Demonstration of the usage of arus.core.Pipeline (3)
====================================================

The pipeline uses multiple sensor streams.
"""

from arus.core.pipeline import Pipeline
import arus
import datetime as dt
import pandas as pd


def _pipeline_test_processor(chunk_list, **kwargs):
    import pandas as pd
    result = {'NAME': [],
              'START_TIME': [], 'STOP_TIME': []}
    for data, st, et, prev_st, prev_et, name in chunk_list:
        result['NAME'].append(name)
        result['START_TIME'].append(data.iloc[0, 0])
        result['STOP_TIME'].append(data.iloc[-1, 0])
    result = pd.DataFrame.from_dict(result)
    return result


if __name__ == "__main__":

    window_size = 12.8
    start_time = dt.datetime.now()

    gr1 = arus.generator.RandomAccelDataGenerator(
        sr=80, grange=8, sigma=1, buffer_size=100)
    seg1 = arus.segmentor.SlidingWindowSegmentor(window_size)
    stream1 = arus.Stream(gr1, seg1, name='stream-1')

    gr2 = arus.generator.RandomAccelDataGenerator(
        sr=50, grange=4, sigma=2, buffer_size=100)
    seg2 = arus.segmentor.SlidingWindowSegmentor(window_size)
    stream2 = arus.Stream(gr2, seg2, name='stream-2')

    pipeline = Pipeline(max_processes=2, scheduler='threads')
    pipeline.add_stream(stream1)
    pipeline.add_stream(stream2)
    pipeline.set_processor(_pipeline_test_processor)
    # for multiple streams, we have to serve start_time to sync streams
    pipeline.start(start_time=start_time)
    results = []
    for result, st, et, prev_st, prev_et, name in pipeline.get_iterator():
        result['WINDOW_ST'] = st
        result['WINDOW_ET'] = et
        result['PREV_WINDOW_ST'] = prev_st
        result['PREV_WINDOW_ET'] = prev_et
        result['STREAM_NAME'] = name
        results.append(result)
        if len(results) == 10:
            break
    pipeline.stop()
    print(pd.concat(results, axis=0, sort=False))



from . import operator
from . import scheduler


class Processor(operator.Operator):
    def __init__(self, func, **kwargs):
        super().__init__()
        self._func = func
        self._scheduler = scheduler.Scheduler(**kwargs)

    def run(self, values=None, src=None, context={}):
        new_context = self.merge_context(context)
        self._scheduler.submit(self._func, values, src=src, **new_context)

    def stop(self):
        super().stop()
        self._scheduler.close()
        self._scheduler.reset()

    def shutdown(self):
        self._scheduler.shutdown()

    def get_result(self):
        try:
            while True:
                if self._stop:
                    break
                result, new_context = self._scheduler.get_result(timeout=0.1)
                yield result, new_context
        except scheduler.Scheduler.ResultNotAvailableError:
            pass



"""
Demonstration of the usage of arus.models.muss.get_inference_pipeline
=====================================================================

The pipeline uses a single sensor generator stream.
"""

from arus.models.muss import MUSSModel
from arus.testing import load_test_data
import pandas as pd
import arus
import datetime as dt


def train_test_classifier(muss):
    feature_filepath, _ = load_test_data(file_type='mhealth', sensor_type='feature',
                                         file_num='single', exception_type='single_placement')
    class_filepath, _ = load_test_data(file_type='mhealth', sensor_type='class_labels',
                                       file_num='single', exception_type='multi_tasks')
    feature_set = pd.read_csv(
        feature_filepath, infer_datetime_format=True, parse_dates=[0, 1, 2])
    feature_set = feature_set[['HEADER_TIME_STAMP',
                               'START_TIME', 'STOP_TIME'] + muss.get_feature_names()]
    class_set = pd.read_csv(
        class_filepath, infer_datetime_format=True, parse_dates=[0, 1, 2])
    class_set = class_set[['HEADER_TIME_STAMP',
                           'START_TIME', 'STOP_TIME', 'MUSS_3_POSTURES']]
    input_feature, input_class = muss.sync_feature_and_class(
        feature_set, class_set)
    filtered_feature, filtered_class = muss.remove_classes(
        input_feature, input_class, class_col='MUSS_3_POSTURES', classes_to_remove=['Unknown', 'Transition'])
    model = muss.train_classifier(
        filtered_feature, filtered_class, class_col='MUSS_3_POSTURES', feature_names=muss.get_feature_names(), placement_names=['DW'])
    return model


def prepare_streams():
    window_size = 12.8
    gr = arus.generator.RandomAccelDataGenerator(
        sr=50, grange=8, sigma=1, buffer_size=100)
    seg = arus.segmentor.SlidingWindowSegmentor(window_size)
    stream = arus.Stream(gr, seg, name='DW')
    return stream


if __name__ == "__main__":
    start_time = dt.datetime.now()
    muss = MUSSModel()
    model = train_test_classifier(muss)
    stream = prepare_streams()
    muss_pipeline = muss.get_inference_pipeline(
        stream, model=model, scheduler='processes', max_processes=2, DW={'sr': 50})
    muss_pipeline.start(start_time=start_time)
    i = 0
    for data, _, _, _, _, name in muss_pipeline.get_iterator():
        i = i + 1
        print(data)
        if i == 5:
            break
    muss_pipeline.stop()
